/* @noflow */

import express from "express"
import * as URL from "url"
import * as Route from "route.flow"
import * as Float from "float.flow"
const rootRoute = Route.Root
const plusRoute = rootRoute
  .segment("calculator")
  .param(Route.Float)
  .segment("+")
  .param(Route.Float)

plusRoute.formatPath(Float.toFloat(313), Float.toFloat(3)) //?

const getRoot = res => {
  const path = plusRoute
    .formatPath(Float.toFloat(313), Float.toFloat(3))
    .toString()
  res.send(`<a href='${path}'>Calculate 313 + 3</a>`)
}

const getPlus = (res, a: number, b: number) => {
  res.send(`${a + b}`)
}

const app = express()
app.use((req, res) => {
  const url = URL.parse(req.url)

  const root = rootRoute.parsePath(url)
  if (root) {
    return getRoot(res, ...root)
  }

  const plus = plusRoute.parsePath(url)
  if (plus) {
    return getPlus(res, ...plus)
  }
})
// app.listen(8090)

Route.String.parsePath({ pathname: "alice" }) //?
Route.String.parsePath({ pathname: "alice/" }) //?
Route.String.parsePath({ pathname: "alice/blog" }) //?
Route.String.parsePath({ pathname: "42" }) //?
Route.String.parsePath({ pathname: "/alice" }) //?

Route.Float.parsePath({ pathname: "42/" }) //?
Route.Float.parsePath({ pathname: "-42.5/" }) //?
Route.Float.parsePath({ pathname: "NaN/" }) //?
Route.Float.parsePath({ pathname: "Infinity/" }) //?
Route.Float.parsePath({ pathname: "Bob/" }) //?

Route.parsePath(Route.Integer, { pathname: "42/" }) //?
Route.parsePath(Route.Integer, { pathname: "-7" }) //?
Route.parsePath(Route.Integer, { pathname: "+8" }) //?
Route.parsePath(Route.Integer, { pathname: "42.2/" }) //?
Route.parsePath(Route.Integer, { pathname: "/" }) //?
Route.parsePath(Route.Integer, { pathname: "Infinity" }) //?
Route.parsePath(Route.Integer, { pathname: "NaN/" }) //?

Route.Root.parsePath({ pathname: "/" }) //?
Route.Root.parsePath({ pathname: "" }) //?
Route.Root.parsePath({ pathname: "/foo" }) //?
Route.Root.parsePath({ pathname: "bar" }) //?

Route.segment("blog").parsePath({ pathname: "blog" }) //?
Route.segment("blog").parsePath({ pathname: "blog/" }) //?
Route.segment("blog").parsePath({ pathname: "blog/cat" }) //?
Route.segment("blog").parsePath({ pathname: "/blog/" }) //?
Route.segment("blog").parsePath({ pathname: "glob" }) //?
Route.segment("blog").parsePath({ pathname: "/" }) //?

const blogID = Route.concat(Route.segment("blog"), Route.Integer)
blogID.parsePath({ pathname: "blog/35/" }) //?
blogID.parsePath({ pathname: "blog/42/" }) //?
blogID.parsePath({ pathname: "blog/" }) //?
blogID.parsePath({ pathname: "42" }) //?

const blogSearch = Route.concat(Route.segment("blog"), Route.segment("search"))
const searchTerm = Route.concat(blogSearch, Route.String)

searchTerm.parsePath({ pathname: "blog/search/cats/" }) //?
searchTerm.parsePath({ pathname: "blog/search/42/" }) //?
searchTerm.parsePath({ pathname: "/search/cats/" }) //?
searchTerm.parsePath({ pathname: "/blog/cats/" }) //?

Route.Root.segment("blog").parsePath({ pathname: "/" }) //?
Route.Root.segment("blog").parsePath({ pathname: "/blog" }) //?

Route.Float.segment("inc").parsePath({ pathname: "cat/inc" }) //?
Route.Float.segment("inc").parsePath({ pathname: "7/inc" }) //?

const blog = Route.concat(Route.Root, Route.segment("blog"))
blog.parsePath({ pathname: "/" }) //?
blog.parsePath({ pathname: "/blog" }) //?

const inc = Route.concat(Route.Float, Route.segment("inc"))
inc.parsePath({ pathname: "cat/inc" }) //?
inc.parsePath({ pathname: "7/inc" }) //?

const calculator = Route.Root
  .segment("calculator")
  .param(Route.Float)
  .segment("+")
  .param(Route.Float)

calculator.parsePath({ pathname: "/calculator/313/+/3" }) //?
calculator.parsePath({ pathname: "/calculator/313/+/" }) //?
calculator.parsePath({ pathname: "/calculator/13/+/4.2/" }) //?

const blogPosts = Route.Root.segment("blog")
const postID = Route.segment("post").param(Route.Integer)
const blogPostID = blogPosts.concat(postID)

blogPostID.parsePath({ pathname: "/blog/post/35/" }) //?
blogPostID.parsePath({ pathname: "/post/42/" }) //?
blogPostID.parsePath({ pathname: "blog/post/7" }) //?
blogPostID.parsePath({ pathname: "/blog/post/" }) //?

const search = Route.concat(Route.segment("blog"), Route.segment("search"))
const term = Route.concat(search, Route.String)

term.parsePath({ pathname: "blog/search/cats/" }) //?
term.parsePath({ pathname: "blog/search/42/" }) //?
term.parsePath({ pathname: "/search/cats/" }) //?
term.parsePath({ pathname: "/blog/cats/" }) //?

const css = Route.param(
  ($: string): ?string => ($.endsWith(".css") ? $ : null),
  String
)

css.parsePath({ pathname: "base.css" }) //?
css.parsePath({ pathname: "fontawesome-webfont.woff2" }) //?
css.parsePath({ pathname: "style/base.css" }) //?

const stylesheet = Route.Root.segment("style").param(css)

stylesheet.parsePath({ pathname: "/style/base.css" }) //?
stylesheet.parsePath({ pathname: "base.css" }) //?
stylesheet.parsePath({ pathname: "style/base.css" }) //?
stylesheet.parsePath({ pathname: "/style/font.woff2" }) //?

const limit = Route.query("limit", Route.Integer)
limit.parsePath({ search: "?limit=5" }) //?
limit.parsePath({ search: "?limit=" }) //?
limit.parsePath({ search: "?limit=0" }) //?
limit.parsePath({ search: "?foo&bar&limit=2" }) //?

const find = Route.segment("find").param(Route.String).concat(limit)

find.parsePath({ search: "?limit=5" }) //?
find.parsePath({ pathname: "find", search: "?limit=5" }) //?
find.parsePath({ pathname: "find/cat", search: "?limit=5" }) //?
find.parsePath({ pathname: "find/cat", search: "?limit=5&sort=asc" }) //?

const seek = Route.Root
  .segment("seek")
  .param(Route.String)
  .query("limit", Route.Integer)

seek.parsePath({ search: "?limit=5" }) //?
seek.parsePath({ pathname: "/seek", search: "?limit=5" }) //?
seek.parsePath({ pathname: "/seek/cat", search: "?limit=5" }) //?

Route.formatPath(Route.segment("find").param(Route.String), "cats") //?

Route.formatPath(
  Route.Root
    .segment("blog")
    .param(Route.String)
    .segment("tag")
    .param(Route.String)
    .segment(),
  "cats",
  "breed"
) //?

Route.segment("find").param(Route.String).formatPath("cats") //?

Route.Root
  .segment("blog")
  .param(Route.String)
  .segment("tag")
  .param(Route.String)
  .segment()
  .formatPath("cats", "breed") //?

Route.formatHash(Route.segment("find").param(Route.String), "cats") //?

Route.formatHash(
  Route.Root
    .segment("blog")
    .param(Route.String)
    .segment("tag")
    .param(Route.String)
    .segment(),
  "cats",
  "breed"
) //?

Route.segment("find").param(Route.String).formatHash("cats") //?

Route.Root
  .segment("blog")
  .param(Route.String)
  .segment("tag")
  .param(Route.String)
  .segment()
  .formatHash("cats", "breed") //?

Route.format(Route.segment("find").param(Route.String), "cats") //?

Route.format(
  Route.Root
    .segment("blog")
    .param(Route.String)
    .segment("tag")
    .param(Route.String)
    .segment(),
  "cats",
  "breed"
) //?

Route.segment("find").param(Route.String).format("cats") //?

Route.Root
  .segment("blog")
  .param(Route.String)
  .segment("tag")
  .param(Route.String)
  .segment()
  .format("cats", "breed") //?

const res = Route.Root.segment("resource").rest(Route.String)
res.parsePath({ pathname: "/resource/foo/bar.js", search: "?foo=bar" }) //?
res.formatPath("hi/is/there/anything.js") //?

Route.segment("find")
  .param(Route.String)
  .segment()
  .parsePath({ pathname: "find/cat/" }) //?
